/* image.cc
 * Copyright (C) 2024 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"
#include "image.hh"
#include <cerrno>
#include <fstream>
#include <sstream>
#include <iostream>
#include <map>
#include <limits>
#include <algorithm>
#include "ab_logging.hh"
#include "options.hh"

#ifdef HAVE_SDL2_IMAGE
#  include <SDL2/SDL.h>
#  include <SDL2/SDL_video.h>
#  include <SDL2/SDL_image.h>
#endif

ImageFP::ImageFP( size_t w, size_t h ) :
    m_width( w ),
    m_height( h )
{
    m_image.clear();
    m_image.resize( w * h );
}

Image444::Image444( size_t w, size_t h ) :
    m_width( w ),
    m_height( h )
{
    m_image.clear();
    m_image.resize( w * h );
    m_marked_pixels.resize( w * h );
}

#ifdef HAVE_SDL2_IMAGE
static void get_rgb( SDL_Surface *surface,
                     size_t x, size_t y,
                     uint8_t &red,
                     uint8_t &green,
                     uint8_t &blue )
{
    uint8_t *pixel = (uint8_t*)surface->pixels;

    pixel += y * surface->pitch;
    pixel += x * surface->format->BytesPerPixel;

     if ( surface->format->BitsPerPixel == 24 ) {
        red = pixel[surface->format->Rshift / 8];
        green = pixel[surface->format->Gshift / 8];
        blue = pixel[surface->format->Bshift / 8];
    } else {
        LOG_CRITICALXX( "texture format not supported" );
        abort();
    }
}
#endif

int ImageFP::read( const std::string &filename )
{
    bool success = false;
#ifdef HAVE_SDL2_IMAGE
    SDL_Surface *surface = IMG_Load( global_options.input_file.c_str() );

    if ( surface ) {
        LOG_DEBUGXX( "got surface" );
        SDL_Surface *ts = SDL_ConvertSurfaceFormat( surface, SDL_PIXELFORMAT_BGR24, 0 );
        if ( ts ) {
            m_width = surface->w;
            m_height = surface->h;

            if ( m_width >= 16384 ||
                 m_height >= 16384 ) {
                std::cerr << "Image too large: "  << m_width << " x " << m_height << std::endl;
                SDL_FreeSurface( ts );
                SDL_FreeSurface( surface );
                return -ERANGE;
            }

            m_image.clear();
            m_image.resize( m_width * m_height );

            for ( size_t y = 0; y < m_height; y++ ) {
                for ( size_t x = 0; x < m_width; x++ ) {
                    uint8_t r, g, b;

                    get_rgb( ts, x, y, r, g, b );

                    PixelFP p( r,
                               g,
                               b,
                               0,
                               256 );

                    m_image.at( y * m_width + x ) = p;
                }
            }
            SDL_FreeSurface( ts );

            success = true;

            LOG_DEBUGXX( "successfully read image" );
        }
        SDL_FreeSurface( surface );
    } else {
        LOG_DEBUGXX( "got no surface" );
    }
#endif

    if ( success ) return 0;

    LOG_DEBUGXX( "try ppm loader" );

    std::ifstream i;
    std::ostringstream o;

    i.open( filename, std::ifstream::binary );

    if ( ! i.good() ) {
        std::cerr << "Error opening file" << std::endl;
        return -EFAULT;
    }

    std::string ppm_format, ppm_size, ppm_color_limit;
    for (; ! i.eof() ;) {
        std::string l;
        std::getline( i, l );

        if ( l.empty() ) continue;
        if ( l[0] == '#' ) continue;

        if ( ppm_format.empty() ) {
            ppm_format = l;
        } else if ( ppm_size.empty() ) {
            ppm_size = l;
        } else if ( ppm_color_limit.empty() ) {
            ppm_color_limit = l;
            break;
        }
    }

    o << i.rdbuf();

    std::string raw = o.str();

    if ( ppm_format != "P6" ) {
        std::cerr << "Unsupported PPM format "  << ppm_format << std::endl;
        return -ENOTSUP;
    }

    if ( sscanf( ppm_size.c_str(), "%lu %lu", &m_width, &m_height ) != 2 ) {
        std::cerr << "Invalid size "  << ppm_size << std::endl;
        return -ERANGE;
    }

    if ( m_width >= 16384 ||
         m_height >= 16384 ) {
        std::cerr << "Image too large: "  << m_width << " x " << m_height << std::endl;
        return -ERANGE;
    }

    m_image.clear();
    m_image.resize( m_width * m_height );

    uint32_t max_color;

    if ( sscanf( ppm_color_limit.c_str(), "%u", &max_color ) != 1 ) {
        std::cerr << "Invalid color limit "  << ppm_color_limit << std::endl;
        return -ERANGE;
    }
    max_color++;

    LOG_DEBUGXX( "Raw size: " << raw.size() );

    for ( size_t y = 0; y < m_height; y++ ) {
        for ( size_t x = 0; x < m_width; x++ ) {
            size_t rawpos = ( y * m_width + x ) * 3;

            if ( rawpos + 2 >= raw.size() ) {
                std::cerr << "Missing pixel data for " << x << "," << y << std::endl;
                break;
            }

            PixelFP p( (uint8_t)raw[rawpos],
                       (uint8_t)raw[rawpos + 1],
                       (uint8_t)raw[rawpos + 2],
                       0,
                       max_color );

            m_image.at( y * m_width + x ) = p;
        }
    }

    return 0;
}

size_t ImageFP::width() const
{
    return m_width;
}

size_t ImageFP::height() const
{
    return m_height;
}
    
size_t Image444::width() const
{
    return m_width;
}

size_t Image444::height() const
{
    return m_height;
}
    
int ImageFP::set_pixel( size_t x, size_t y, PixelFP v )
{
    if ( x >= m_width ) return -ERANGE;
    if ( y >= m_height ) return -ERANGE;

    m_image[ y * m_width + x ] = v;

    return 0;
}

int Image444::set_pixel( size_t x, size_t y, Pixel444 v )
{
    if ( x >= m_width ) return -ERANGE;
    if ( y >= m_height ) return -ERANGE;

    m_image[ y * m_width + x ] = v;

    return 0;
}

std::optional< PixelFP > ImageFP::get_pixel( size_t x, size_t y ) const
{
    if ( x >= m_width ) return {};
    if ( y >= m_height ) return {};

    return m_image[ y * m_width + x ];
}

std::optional< Pixel444 > Image444::get_pixel( size_t x, size_t y ) const
{
    if ( x >= m_width ) return {};
    if ( y >= m_height ) return {};

    return m_image[ y * m_width + x ];
}

double ImageFP::calc_error( const ImageFP &other ) const
{
    if ( other.m_width != m_width ) return std::numeric_limits<double>::infinity();
    if ( other.m_height != m_height ) return std::numeric_limits<double>::infinity();

    double e = 0.0;
    
    for ( size_t y = 0; y < m_height; y++ ) {
        for ( size_t x = 0; x < m_width; x++ ) {
            auto my_pixel = get_pixel( x, y );
            auto oth_pixel = other.get_pixel( x, y );

            if ( ! my_pixel ) return std::numeric_limits<double>::infinity();
            if ( ! oth_pixel ) return std::numeric_limits<double>::infinity();

#if 1
            double d = my_pixel->r - oth_pixel->r;
            e += d * d;
            d = my_pixel->g - oth_pixel->g;
            e += d * d;
            d = my_pixel->b - oth_pixel->b;
            e += d * d;
#else
            e += my_pixel->distance( *oth_pixel, PIXEL_FLOAT );
#endif
        }
    }

    return e;
}

double Image444::calc_error( const Image444 &other ) const
{
    if ( other.m_width != m_width ) return std::numeric_limits<double>::infinity();
    if ( other.m_height != m_height ) return std::numeric_limits<double>::infinity();

    double e = 0.0;
    
    for ( size_t y = 0; y < m_height; y++ ) {
        for ( size_t x = 0; x < m_width; x++ ) {
            auto my_pixel = get_pixel( x, y ).value_or( Pixel444() );
            auto oth_pixel = other.get_pixel( x, y ).value_or( Pixel444() );

            //if ( ! my_pixel ) return std::numeric_limits<double>::infinity();
            //if ( ! oth_pixel ) return std::numeric_limits<double>::infinity();

            auto myfp = my_pixel.convert();
            auto othfp = oth_pixel.convert();
#if 1
            double d = myfp.r - othfp.r;
            e += d * d;
            d = myfp.g - othfp.g;
            e += d * d;
            d = myfp.b - othfp.b;
            e += d * d;
#else
            e += my_pixel->distance( *oth_pixel, PIXEL_FLOAT );
#endif
        }
    }

    return e;
}

double Image444::calc_error( const ImageFP &other ) const
{
    if ( other.width() != m_width ) return std::numeric_limits<double>::infinity();
    if ( other.height() != m_height ) return std::numeric_limits<double>::infinity();

    double e = 0.0;
    
    for ( size_t y = 0; y < m_height; y++ ) {
        for ( size_t x = 0; x < m_width; x++ ) {
            auto my_pixel = get_pixel( x, y ).value_or( Pixel444() );
            auto oth_pixel = other.get_pixel( x, y ).value_or( PixelFP() );

            //if ( ! my_pixel ) return std::numeric_limits<double>::infinity();
            //if ( ! oth_pixel ) return std::numeric_limits<double>::infinity();

            auto myfp = my_pixel.convert();
#if 1
            double d = myfp.r - oth_pixel.r;
            e += d * d;
            d = myfp.g - oth_pixel.g;
            e += d * d;
            d = myfp.b - oth_pixel.b;
            e += d * d;
#else
            e += my_pixel->distance( *oth_pixel, PIXEL_FLOAT );
#endif
        }
    }

    return e;
}

bool Image444::is_ham_neighbor( const Pixel444 &left_pixel,
                                const Pixel444 &right_pixel ) const
{
    if ( left_pixel.r == right_pixel.r &&
         left_pixel.g == right_pixel.g ) return true;

    if ( left_pixel.r == right_pixel.r &&
         left_pixel.b == right_pixel.b ) return true;

    if ( left_pixel.g == right_pixel.g &&
         left_pixel.b == right_pixel.b ) return true;

    return false;
}

bool Image444::is_ham() const
{
    if ( m_palette.size() > 16 ) return false;

    for ( size_t y = 0; y < m_height; y++ ) {
        for ( size_t x = 0; x < m_width; x++ ) {
            auto pixel = get_pixel( x, y );
            if ( ! pixel ) return false;

            if ( m_palette.find_color( *pixel ) ) {
                continue;
            }

            // first pixel of a row must be a palette entry
            if ( x == 0 ) return false;

            auto left_pixel = get_pixel( x - 1, y );
            if ( ! left_pixel ) return false;

            if ( is_ham_neighbor( *left_pixel, *pixel ) ) continue;

            return false;
        }
    }

    return true;
}

bool Image444::write_ppm( const std::string &filename ) const
{
    std::ofstream ofile( filename, std::ofstream::binary );
    if ( ! ofile.is_open() ) {
        return false;
    }

    ofile << "P6\n";
    ofile << m_width << " " << m_height << "\n";
    ofile << "255\n";

    for ( size_t y = 0; y < m_height; y++ ) {
        for ( size_t x = 0; x < m_width; x++ ) {
            auto pixel = get_pixel( x, y ).value_or( Pixel444() );

            uint8_t r = 0;
            uint8_t g = 0;
            uint8_t b = 0;

            r = ( pixel.r << 4 ) | pixel.r;
            g = ( pixel.g << 4 ) | pixel.g;
            b = ( pixel.b << 4 ) | pixel.b;

            char buf[3] = { (char)r, (char)g ,(char)b };
            ofile.write( buf, sizeof( buf ) );
        }
    }

    return true;
}

Image444 ImageFP::convert_to_444() const
{
    Image444 img( m_width, m_height );

    for ( size_t y = 0; y < m_height; y++ ) {
        for ( size_t x = 0; x < m_width; x++ ) {
            auto &fppixel = m_image.at( y * m_width + x );

            Pixel444 ocs_pixel = fppixel.convert();
            img.set_pixel( x, y, ocs_pixel );
        }
    }

    return img;
}

std::vector< std::pair< Pixel444, size_t > > Image444::get_unique_colors( bool skip_horizontal_continuous ) const
{
    std::map< Pixel444, size_t > count;

    for ( size_t y = 0; y < m_height; y++ ) {
        for ( size_t x = 0; x < m_width; x++ ) {
            if ( m_marked_pixels[ y * m_width + x ] ) continue;

            auto &pixel = m_image.at( y * m_width + x );

            if ( skip_horizontal_continuous && x > 0 ) {
                auto &left_pixel = m_image.at( y * m_width + x - 1 );
                if ( pixel == left_pixel ) continue;
            }

            if ( count.count( pixel ) == 0 ) {
                count[pixel] = 1;
            } else {
                count[pixel]++;
            }
        }
    }

    std::vector< std::pair< Pixel444, size_t > > res;

    for ( auto &e : count ) {
        res.push_back( e );
    }

    std::sort( res.begin(),
               res.end(), []( const auto &lhs, const auto &rhs ) {
                   return lhs.second > rhs.second;
               });

    return res;
}

size_t Image444::calc_unique_colors() const
{
    auto pal = get_unique_colors( false );

    return pal.size();
}

void Image444::set_palette( const Palette &palette )
{
    m_palette = palette;
}

const Palette &Image444::get_palette() const
{
    return m_palette;
}

int Image444::palettize()
{
    for ( size_t y = 0; y < m_height; y++ ) {
        for ( size_t x = 0; x < m_width; x++ ) {
            auto &pixel = m_image[ y * m_width + x ];
            auto pal_entry = m_palette.find_nearest_color( pixel );
            if ( ! pal_entry ) {
                return -EINVAL;
            }

            pixel = *m_palette.get_color( *pal_entry );
        }
    }

    return 0;
}

size_t Image444::count_color_ham_neighbor( Pixel444 v ) const
{
    size_t c = 0;

    //std::cout << "  check pixel " << v << std::endl;

    for ( size_t y = 0; y < m_height; y++ ) {
        for ( size_t x = 0; x < m_width; x++ ) {
            auto pixel = get_pixel( x, y );
            if ( ! pixel ) continue;

            if ( ! ( *pixel == v ) ) continue;
            if ( x + 1 >= m_width ) {
                c++;
                continue;
            }

            auto right_pixel = get_pixel( x + 1, y );
            if ( ! pixel ) continue;

            if ( is_ham_neighbor( *pixel, *right_pixel ) ) {
                c++;
            }
        }
    }

    return c;
}

bool Image444::is_not_ham_surrounded( const std::optional< Pixel444 > &left_pixel,
                                      const std::optional< Pixel444 > &mid_pixel,
                                      const std::optional< Pixel444 > &right_pixel ) const
{
    if ( ! left_pixel ) {
        if ( is_ham_neighbor( *mid_pixel, *right_pixel ) ) return false;

        return true;
    }

    if ( ! right_pixel ) {
        if ( is_ham_neighbor( *left_pixel, *mid_pixel ) ) return false;

        return true;
    }

    if ( is_ham_neighbor( *left_pixel, *mid_pixel ) ) return false;
    if ( is_ham_neighbor( *mid_pixel, *right_pixel ) ) return false;

    return true;
}

std::vector< Pixel444 > Image444::calc_palette( size_t colors, float err )
{
    auto unique_colors = get_unique_colors( true );

    return calc_palette( unique_colors, colors, err );
}

std::vector< Pixel444 > Image444::calc_palette( const std::vector< std::pair< Pixel444, size_t > > &unique_colors,
                                                size_t colors, float err )
{
    if ( colors < 1 ) return {};
    if ( unique_colors.empty() ) return {};

    std::vector< Pixel444 > res;
    std::vector< std::pair< Pixel444, size_t > > remaining = unique_colors;
    std::vector< std::pair< Pixel444, size_t > > remaining_next;

    while ( ! remaining.empty() && res.size() < colors ) {
        PixelFP selected = remaining[0].first.convert();
        res.push_back( selected.convert() );

        remaining_next.clear();

        for ( auto &c : remaining ) {
            if ( selected.distance( c.first.convert() ) > err ) {
                remaining_next.push_back( c );
            }
        }

        remaining = remaining_next;
    }

    return res;
}

void Image444::reset_marks()
{
    m_marked_pixels.clear();
    m_marked_pixels.resize( m_width * m_height );
}

void Image444::simple_hamify()
{
    reset_marks();

    size_t matches = 0, misses = 0;

    for ( size_t y = 0; y < m_height; y++ ) {
        for ( size_t x = 0; x < m_width; x++ ) {
            auto &pixel = m_image[ y * m_width + x ];
            auto pal_entry = m_palette.find_color( pixel );

            if ( pal_entry ) {
                //pixel = Pixel( 0.0 );
                //pixel.ocs_pal.pal_entry = *pal_entry;
                matches++;
                continue;
            }

            misses++;

            auto nearest_entry = m_palette.find_nearest_color( pixel );
            if ( ! nearest_entry ) {
                LOG_CRITICALXX( "failed to find nearest color" );
                abort();
                return;
            }

            if ( x == 0 ) {
                //pixel = Pixel( 0.0 );
                //pixel.ocs_pal.pal_entry = *nearest_entry;
                pixel = *m_palette.get_color( *nearest_entry );
                continue;
            }
            
            auto &previous_pixel = m_image[ y * m_width + x - 1 ];

            if ( global_options.lookahead_mode != ham_options::NO_LOOKAHEAD_FAST ) {
                // only use one pixel lookahead, two pixel seems to make it worse again
                auto [ new_pixel, new_pixel_error ] = ham_step_lookahead( pixel, previous_pixel, get_pixel( x + 1, y ), {}, x, y );
                pixel = new_pixel;
            } else {
                auto [ ham_pixel, ham_value ] = ham_step( pixel, previous_pixel );

                if ( m_palette.get_color( *nearest_entry )->distance( pixel ) <
                     ham_pixel.distance( pixel ) ) {
                    pixel = *m_palette.get_color( *nearest_entry );
                } else {
                    //pixel = Pixel( 0.0 );
                    //pixel.ocs_pal.pal_entry = ham_value;
                    pixel = ham_pixel;
                }
            }
        }
    }

    LOG_DEBUGXX( "  matches: " << matches << ", misses: " << misses );
}

std::pair< Pixel444, uint8_t > Image444::ham_step( const Pixel444 &dst,
                                                   const Pixel444 &src )
{
    int red_distance = abs( (int)src.r - dst.r );
    int green_distance = abs( (int)src.g - dst.g );
    int blue_distance = abs( (int)src.b - dst.b );
    uint8_t ham_value = 1 << 4; // no change

    Pixel444 np = src;

    if ( red_distance >= green_distance &&
         red_distance >= blue_distance ) {
        np.r = dst.r;
        ham_value = ( 2 << 4 ) | dst.r;
    } else if ( green_distance >= red_distance &&
                green_distance >= blue_distance ) {
        np.g = dst.g;
        ham_value = ( 3 << 4 ) | dst.g;
    } else {
        np.b = dst.b;
        ham_value = ( 1 << 4 ) | dst.b;
    }

    return { np, ham_value };
}

std::pair< Pixel444, double > Image444::ham_step_lookahead( const Pixel444 &target,
                                                            const Pixel444 &start,
                                                            const std::optional< Pixel444 > &right,
                                                            const std::optional< Pixel444 > &right_right,
                                                            size_t x, size_t y )
{
    auto target_nearest_entry = m_palette.find_nearest_color( target );
    if ( ! target_nearest_entry ) {
        LOG_CRITICALXX( "failed to find nearest color" );
        abort();
    }
    const Pixel444 target_nearest = *m_palette.get_color( *target_nearest_entry );

    LOG_DEBUGXX( "called for " << start << " -> " << target << ":" );

    Pixel444 red_np = start;
    Pixel444 green_np = start;
    Pixel444 blue_np = start;
    red_np.r = target.r;
    green_np.g = target.g;
    blue_np.b = target.b;

    if ( red_np == target ) return { red_np, 0.0 };
    if ( green_np == target ) return { green_np, 0.0 };
    if ( blue_np == target ) return { blue_np, 0.0 };
    if ( target_nearest == target ) return { target_nearest, 0.0 };

    double red_distance = red_np.distance( target );
    double green_distance = green_np.distance( target );
    double blue_distance = blue_np.distance( target );
    double nearest_distance = target_nearest.distance( target );

    // double red_next_distance = 0.0;
    // double green_next_distance = 0.0;
    // double blue_next_distance = 0.0;
    // double nearest_next_distance = 0.0;
    double red_next_ham_error = 0.0;
    double green_next_ham_error = 0.0;
    double blue_next_ham_error = 0.0;
    double nearest_next_ham_error = 0.0;

    if ( right && global_options.lookahead_mode != ham_options::NO_LOOKAHEAD_NORMAL ) {
        auto [ red_next_ham, _red_next_ham_error ] = ham_step_lookahead( *right, red_np, right_right, {}, x + 1, y );
        auto [ green_next_ham, _green_next_ham_error ] = ham_step_lookahead( *right, green_np, right_right, {}, x + 1, y );
        auto [ blue_next_ham, _blue_next_ham_error ] = ham_step_lookahead( *right, blue_np, right_right, {}, x + 1, y );
        auto [ nearest_next_ham, _nearest_next_ham_error ] = ham_step_lookahead( *right, target_nearest, right_right, {}, x + 1, y );

        // red_next_distance = red_next_ham.distance( *right, m_pixel_type );
        // green_next_distance = green_next_ham.distance( *right, m_pixel_type );
        // blue_next_distance = blue_next_ham.distance( *right, m_pixel_type );
        // nearest_next_distance = nearest_next_ham.distance( *right, m_pixel_type );

        red_next_ham_error = _red_next_ham_error;
        green_next_ham_error = _green_next_ham_error;
        blue_next_ham_error = _blue_next_ham_error;
        nearest_next_ham_error = _nearest_next_ham_error;
    }

    struct pixel_info {
        double direct_error;
        int component;
        double indirect_error;
        double above_error;
    };

    std::vector< struct pixel_info > errors = { { red_distance + red_next_ham_error, 0, red_next_ham_error, 0.0 },
                                                { green_distance + green_next_ham_error, 1, green_next_ham_error, 0.0 },
                                                { blue_distance + blue_next_ham_error, 2, blue_next_ham_error, 0.0 },
                                                { nearest_distance + nearest_next_ham_error, 3, nearest_next_ham_error, 0.0 } };
    std::sort( errors.begin(), errors.end(),
               []( const auto &lhs, const auto &rhs ) {
                   if ( lhs.direct_error == rhs.direct_error ) return lhs.indirect_error < rhs.indirect_error;
                   return lhs.direct_error < rhs.direct_error;
               } );

    if ( errors[0].direct_error == errors[1].direct_error &&
         errors[0].indirect_error == errors[1].indirect_error &&
         global_options.lookahead_mode == ham_options::DIRECT_LOOKAHEAD ) {
        // if the first two entries have the same error, look above
        // the last two lines to make decision based on that. It can
        // be worse in non-dither mode if we just use the line above
        // as it repeats the same error again make it harder to make a
        // different decision. In this case it can help to just use
        // the line before the previous line but using both seems good
        // too.
        double red_above_error = 0.0;
        double green_above_error = 0.0;
        double blue_above_error = 0.0;
        double nearest_above_error = 0.0;

        red_above_error = get_pixel( x - 1, y - 2).value_or( Pixel444() ).distance( red_np ) * 0.25 +
            get_pixel( x, y - 2).value_or( Pixel444() ).distance( red_np ) * 0.5 +
            get_pixel( x + 1, y - 2).value_or( Pixel444() ).distance( red_np ) * 0.25;
        green_above_error = get_pixel( x - 1, y - 2).value_or( Pixel444() ).distance( green_np ) * 0.25 +
            get_pixel( x, y - 2).value_or( Pixel444() ).distance( green_np ) * 0.5 +
            get_pixel( x + 1, y - 2).value_or( Pixel444() ).distance( green_np ) * 0.25;
        blue_above_error = get_pixel( x - 1, y - 2).value_or( Pixel444() ).distance( blue_np ) * 0.25 +
            get_pixel( x, y - 2).value_or( Pixel444() ).distance( blue_np ) * 0.5 +
            get_pixel( x + 1, y - 2).value_or( Pixel444() ).distance( blue_np ) * 0.25;
        nearest_above_error = get_pixel( x - 1, y - 2).value_or( Pixel444() ).distance( target_nearest ) * 0.25 +
            get_pixel( x, y - 2).value_or( Pixel444() ).distance( target_nearest ) * 0.5 +
            get_pixel( x + 1, y - 2).value_or( Pixel444() ).distance( target_nearest ) * 0.25;

        red_above_error += get_pixel( x - 1, y - 1).value_or( Pixel444() ).distance( red_np ) * 0.25 +
            get_pixel( x, y - 1).value_or( Pixel444() ).distance( red_np ) * 0.5 +
            get_pixel( x + 1, y - 1).value_or( Pixel444() ).distance( red_np ) * 0.25;
        green_above_error += get_pixel( x - 1, y - 1).value_or( Pixel444() ).distance( green_np ) * 0.25 +
            get_pixel( x, y - 1).value_or( Pixel444() ).distance( green_np ) * 0.5 +
            get_pixel( x + 1, y - 1).value_or( Pixel444() ).distance( green_np ) * 0.25;
        blue_above_error += get_pixel( x - 1, y - 1).value_or( Pixel444() ).distance( blue_np ) * 0.25 +
            get_pixel( x, y - 1).value_or( Pixel444() ).distance( blue_np ) * 0.5 +
            get_pixel( x + 1, y - 1).value_or( Pixel444() ).distance( blue_np ) * 0.25;
        nearest_above_error += get_pixel( x - 1, y - 1).value_or( Pixel444() ).distance( target_nearest ) * 0.25 +
            get_pixel( x, y - 1).value_or( Pixel444() ).distance( target_nearest ) * 0.5 +
            get_pixel( x + 1, y - 1).value_or( Pixel444() ).distance( target_nearest ) * 0.25;

        errors = { { red_distance + red_next_ham_error, 0, red_next_ham_error, red_above_error },
                   { green_distance + green_next_ham_error, 1, green_next_ham_error, green_above_error },
                   { blue_distance + blue_next_ham_error, 2, blue_next_ham_error, blue_above_error },
                   { nearest_distance + nearest_next_ham_error, 3, nearest_next_ham_error, nearest_above_error } };

        std::sort( errors.begin(), errors.end(),
                   []( const auto &lhs, const auto &rhs ) {
                       if ( lhs.direct_error == rhs.direct_error ) {
                           if ( lhs.indirect_error == rhs.indirect_error ) return lhs.above_error < rhs.above_error;
                           return lhs.indirect_error < rhs.indirect_error;
                       }
                       return lhs.direct_error < rhs.direct_error;
                   } );
    }

    switch( errors[0].component ) {
        case 0:
            return { red_np, errors[0].direct_error };
        case 1:
            return { green_np, errors[0].direct_error };
        case 2:
            return { blue_np, errors[0].direct_error };
        default:
            return { target_nearest, errors[0].direct_error };
    }
}

std::optional< std::pair< Pixel444, uint8_t > > Image444::get_ham_pixel( size_t x, size_t y ) const
{
    if ( x >= m_width || y >= m_height ) return {};

    auto &pixel = m_image[ y * m_width + x ];

    auto pal_entry = m_palette.find_color( pixel );

    if ( pal_entry ) {
        size_t entry = *pal_entry;

        if ( entry >= 16 ) {
            LOG_CRITICALXX( "not a 16 color palette!" );
            return {};
        }

        return { { pixel, entry } };
    }

    if ( x == 0 ) return {};

    auto previous_pixel_pair = get_ham_pixel( x - 1, y );
    if ( ! previous_pixel_pair ) return {};

    auto & [ previous_pixel, previous_ham_value ] = *previous_pixel_pair;

    uint8_t ham_value;

    if ( pixel.r != previous_pixel.r ) {
        ham_value = ( 2 << 4 ) | pixel.r;

        if ( pixel.g != previous_pixel.g ||
             pixel.b != previous_pixel.b ) {
            LOG_CRITICALXX( "not a ham image" );
            return {};
        }
    } else if ( pixel.g != previous_pixel.g ) {
        ham_value = ( 3 << 4 ) | pixel.g;

        if ( pixel.r != previous_pixel.r ||
             pixel.b != previous_pixel.b ) {
            LOG_CRITICALXX( "not a ham image" );
            return {};
        }
    } else {
        ham_value = ( 1 << 4 ) | pixel.b;

        if ( pixel.r != previous_pixel.r ||
             pixel.g != previous_pixel.g ) {
            LOG_CRITICALXX( "not a ham image" );
            return {};
        }
    }

    if ( ham_value >= 64 ) abort();
    return { { pixel, ham_value } };
}

static double interpolate( double l, double r, double f )
{
    return f * r - ( f - 1 ) * l;
}

ImageFP ImageFP::scale( size_t w, size_t h ) const
{
    ImageFP res( w, h );

    if ( m_width < 1 || m_height < 1 ) return res;

    for ( size_t y = 0; y < h; y++ ) {
        for ( size_t x = 0; x < w; x++ ) {
            // since we use pixel (x,y) to (x+1,y+1), scale to
            // width-1,height-1
            double start_x = (double)x * ( m_width - 1 ) / w;
            double end_x = (double)( x + 1 ) * ( m_width - 1 ) / w;

            double start_y = (double)y * ( m_height - 1 ) / h;
            double end_y = (double)( y + 1 ) * ( m_height - 1 ) / h;

            double distance_x = end_x - start_x;
            double distance_y = end_y - start_y;

            //LOG_VERBOSEXX( "scaling for " << x << "," << y << " from " << start_x << "," << start_y << " to " << end_x << "," << end_y );
            //LOG_VERBOSEXX( "   distance: " << distance_x << "," << distance_y );

            double sum_r = 0.0;
            double sum_g = 0.0;
            double sum_b = 0.0;
            double sum_a = 0.0;
            double count = 0;

            size_t current_iy = floor( start_y );
            size_t current_ix = floor( start_x );

            //LOG_VERBOSEXX( "  use pixel " << current_ix << " with weight " << weight_x << "," << weight_y );
            //LOG_VERBOSEXX( "    " << current_ix << "," << start_x );

            double current_y = start_y;
            while ( distance_y > 0.0 ) {
                current_iy = floor( current_y );

                // distance is basically the distance between the
                // target pixel with fraction and the next full
                // pixel. It is used to weight the color value.
                double pixel_distance_y;

                if ( floor( current_y + 1.0 ) > end_y ) {
                    pixel_distance_y = end_y - current_y;
                } else {
                    pixel_distance_y = floor( current_y + 1.0 ) - current_y;
                }

                const double color_value_weight_y = pixel_distance_y;

                // interpolation weight for a single pixel, basically
                // the fraction part weighted based on absolute
                // coordinate so that left pixel are interpolated to
                // the left while right pixel get interpolated to the
                // right
                const double interpolation_weight_y = ( current_y - floor( current_y ) ) + pixel_distance_y * current_y / m_height;

                double current_x = start_x;
                double remaining_distance_x = distance_x;
                while ( remaining_distance_x > 0.0 ) {
                    current_ix = floor( current_x );

                    const PixelFP tl = get_pixel( current_ix, current_iy ).value_or( PixelFP() );
                    const PixelFP tr = get_pixel( current_ix + 1, current_iy ).value_or( PixelFP() );
                    const PixelFP bl = get_pixel( current_ix, current_iy + 1 ).value_or( PixelFP() );
                    const PixelFP br = get_pixel( current_ix + 1, current_iy + 1 ).value_or( PixelFP() );

                    double pixel_distance_x;

                    if ( floor( current_x + 1.0 ) > end_x ) {
                        pixel_distance_x = end_x - current_x;
                    } else {
                        pixel_distance_x = floor( current_x + 1.0 ) - current_x;
                    }

                    const double color_value_weight_x = pixel_distance_x;
                    const double interpolation_weight_x = ( current_x - floor( current_x ) ) + pixel_distance_x * current_x / m_width;

                    double ys = interpolate( tl.r, bl.r, interpolation_weight_y );
                    double ye = interpolate( tr.r, br.r, interpolation_weight_y );
                    double v = interpolate( ys, ye, interpolation_weight_x );

                    sum_r += v * color_value_weight_x * color_value_weight_y;

                    ys = interpolate( tl.g, bl.g, interpolation_weight_y );
                    ye = interpolate( tr.g, br.g, interpolation_weight_y );
                    v = interpolate( ys, ye, interpolation_weight_x );

                    sum_g += v * color_value_weight_x * color_value_weight_y;
                    
                    ys = interpolate( tl.b, bl.b, interpolation_weight_y );
                    ye = interpolate( tr.b, br.b, interpolation_weight_y );
                    v = interpolate( ys, ye, interpolation_weight_x );

                    sum_b += v * color_value_weight_x * color_value_weight_y;

                    count += color_value_weight_x * color_value_weight_y;

                    remaining_distance_x -= pixel_distance_x;
                    current_x = floor( current_x + 1.0 );
                }

                distance_y -= pixel_distance_y;
                current_y  = floor( current_y + 1.0 );
            }

            res.set_pixel( x, y, PixelFP( sum_r / count,
                                          sum_g / count,
                                          sum_b / count,
                                          sum_a / count ) );
        }
    }

    return res;
}

ImageFP ImageFP::dither_444() const
{
    ImageFP img( m_width, m_height );

    for ( size_t y = 0; y < m_height; y++ ) {
        for ( size_t x = 0; x < m_width; x++ ) {
            auto &fppixel = m_image.at( y * m_width + x );
            img.set_pixel( x, y, fppixel );
        }
    }

    for ( size_t y = 0; y < m_height; y++ ) {
        for ( size_t x = 0; x < m_width; x++ ) {
            auto fppixel = img.get_pixel( x, y );

            // only propagate quantisation error, not clipping errors
            if ( fppixel->r > 1.0 ) fppixel->r = 16.0 / 16;
            if ( fppixel->g > 1.0 ) fppixel->g = 16.0 / 16;
            if ( fppixel->b > 1.0 ) fppixel->b = 16.0 / 16;
            Pixel444 ocs_pixel = fppixel->convert();
            PixelFP new_fp_pixel = ocs_pixel.convert();

            img.set_pixel( x, y, new_fp_pixel );

            double red_error = fppixel->r - new_fp_pixel.r;
            double green_error = fppixel->g - new_fp_pixel.g;
            double blue_error = fppixel->b - new_fp_pixel.b;

            if ( auto right = img.get_pixel( x + 1, y ) ) {
                img.set_pixel( x + 1, y, PixelFP( right->r + red_error * 7 / 16,
                                                  right->g + green_error * 7 / 16,
                                                  right->b + blue_error * 7 / 16,
                                                  right->a ) );
            }
            if ( auto left_below = img.get_pixel( x - 1, y + 1 ) ) {
                img.set_pixel( x - 1, y + 1, PixelFP( left_below->r + red_error * 3 / 16,
                                                      left_below->g + green_error * 3 / 16,
                                                      left_below->b + blue_error * 3 / 16,
                                                      left_below->a ) );
            }
            if ( auto below = img.get_pixel( x, y + 1 ) ) {
                img.set_pixel( x, y + 1, PixelFP( below->r + red_error * 5 / 16,
                                                  below->g + green_error * 5 / 16,
                                                  below->b + blue_error * 5 / 16,
                                                  below->a ) );
            }
            if ( auto right_below = img.get_pixel( x + 1, y + 1 ) ) {
                img.set_pixel( x + 1, y + 1, PixelFP( right_below->r + red_error * 1 / 16,
                                                      right_below->g + green_error * 1 / 16,
                                                      right_below->b + blue_error * 1 / 16,
                                                      right_below->a ) );
            }
        }
    }

    return img;
}
