/* image.hh
 * Copyright (C) 2024 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef IMAGE_HH
#define IMAGE_HH

#include "pixel.hh"
#include "palette.hh"
#include <vector>
#include <optional>
#include <string>

class Image444;

class ImageFP {
public:
    ImageFP() = default;
    ImageFP( size_t w, size_t h );

    int read( const std::string &filename );

    size_t width() const;
    size_t height() const;
    
    int set_pixel( size_t x, size_t y, PixelFP v );
    std::optional< PixelFP > get_pixel( size_t x, size_t y ) const;

    double calc_error( const ImageFP &other ) const;

    Image444 convert_to_444() const;

    ImageFP scale( size_t w, size_t h ) const;

    ImageFP dither_444() const;
private:
    size_t m_width = 0;
    size_t m_height = 0;
    std::vector< PixelFP > m_image;
};

class Image444 {
public:
    Image444( size_t w, size_t h );

    size_t width() const;
    size_t height() const;
    
    int set_pixel( size_t x, size_t y, Pixel444 v );
    std::optional< Pixel444 > get_pixel( size_t x, size_t y ) const;

    double calc_error( const Image444 &other ) const;
    double calc_error( const ImageFP &other ) const;

    bool is_ham() const;

    bool write_ppm( const std::string &filename ) const;

    size_t calc_unique_colors() const;

    std::vector< std::pair< Pixel444, size_t > > get_unique_colors( bool skip_horizontal_continuous ) const;
    size_t count_color_ham_neighbor( Pixel444 v ) const;

    void set_palette( const Palette &palette );
    const Palette &get_palette() const;
    int palettize();

    bool is_ham_neighbor( const Pixel444 &left_pixel,
                          const Pixel444 &right_pixel ) const;
    bool is_not_ham_surrounded( const std::optional< Pixel444 > &left_pixel,
                                const std::optional< Pixel444 > &mid_pixel,
                                const std::optional< Pixel444 > &right_pixel ) const;

    std::vector< Pixel444 > calc_palette( size_t colors, float err );
    std::vector< Pixel444 > calc_palette( const std::vector< std::pair< Pixel444, size_t > > &unique_colors,
                                          size_t colors, float err );
    

    void reset_marks();

    void simple_hamify();

    std::pair< Pixel444, uint8_t > ham_step( const Pixel444 &dst, const Pixel444 &src );
    std::pair< Pixel444, double > ham_step_lookahead( const Pixel444 &target,
                                                      const Pixel444 &start,
                                                      const std::optional< Pixel444 > &right,
                                                      const std::optional< Pixel444 > &right_right,
                                                      size_t x, size_t y );

    std::optional< std::pair< Pixel444, uint8_t > > get_ham_pixel( size_t x, size_t y ) const;
private:
    size_t m_width = 0;
    size_t m_height = 0;
    std::vector< Pixel444 > m_image;
    std::vector< bool > m_marked_pixels;
    Palette m_palette;
};

#endif /* IMAGE_HH */
