/* lowlevel_types.cc
 * Copyright (C) 2020-2023 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "lowlevel_types.hh"
#include <endian.h>
#include <cstdint>
#include <cstring>
#include <cstdlib>

uint8_t get_u8( const void *ptr, size_t offset )
{
    const uint8_t *u8ptr = (const uint8_t*)ptr;
    u8ptr += offset;
    return *((const uint8_t*)u8ptr);
}

uint16_t get_u16( const void *ptr, size_t offset )
{
    const uint8_t *u8ptr = (const uint8_t*)ptr;
    u8ptr += offset;

    const std::uintptr_t p = reinterpret_cast< const std::uintptr_t >( u8ptr );
    if ( ( p & 0x1 ) == 0 ) {
        // 16 bit aligned
        const uint16_t *u16p = reinterpret_cast< const uint16_t * >( p );
        return *u16p;
    }

    uint16_t v;

    memcpy( &v, u8ptr, sizeof( v ) );

    return v;
}

uint32_t get_u32( const void *ptr, size_t offset )
{
    const uint8_t *u8ptr = (const uint8_t*)ptr;
    u8ptr += offset;

    const std::uintptr_t p = reinterpret_cast< const std::uintptr_t >( u8ptr );
    if ( ( p & 0x3 ) == 0 ) {
        // 32 bit aligned
        const uint32_t *u32p = reinterpret_cast< const uint32_t * >( p );
        return *u32p;
    }

    uint32_t v;

    memcpy( &v, u8ptr, sizeof( v ) );
    
    return v;
}

uint64_t get_u64( const void *ptr, size_t offset )
{
    const uint8_t *u8ptr = (const uint8_t*)ptr;
    u8ptr += offset;

    const std::uintptr_t p = reinterpret_cast< const std::uintptr_t >( u8ptr );
    if ( ( p & 0x7 ) == 0 ) {
        // 64 bit aligned
        const uint64_t *u64p = reinterpret_cast< const uint64_t * >( p );
        return *u64p;
    }

    uint64_t v;

    memcpy( &v, u8ptr, sizeof( v ) );
    
    return v;
}

int8_t get_s8( const void *ptr, size_t offset )
{
    const uint8_t *u8ptr = (const uint8_t*)ptr;
    u8ptr += offset;
    return *((const int8_t*)u8ptr);
}

int16_t get_s16( const void *ptr, size_t offset )
{
    const uint8_t *u8ptr = (const uint8_t*)ptr;
    u8ptr += offset;
    return *((const int16_t*)u8ptr);
}

int32_t get_s32( const void *ptr, size_t offset )
{
    const uint8_t *u8ptr = (const uint8_t*)ptr;
    u8ptr += offset;
    return *((const int32_t*)u8ptr);
}

int64_t get_s64( const void *ptr, size_t offset )
{
    const uint8_t *u8ptr = (const uint8_t*)ptr;
    u8ptr += offset;
    return *((const int64_t*)u8ptr);
}

uint16_t betoh_u16( uint16_t v )
{
    return be16toh( v );
}

uint32_t betoh_u32( uint32_t v )
{
    return be32toh( v );
}

uint64_t betoh_u64( uint64_t v )
{
    return be64toh( v );
}

uint16_t htobe_u16( uint16_t v )
{
    return htobe16( v );
}

uint32_t htobe_u32( uint32_t v )
{
    return htobe32( v );
}

uint64_t htobe_u64( uint64_t v )
{
    return htobe64( v );
}

void put_u8( void *ptr, size_t offset, uint8_t v )
{
    uint8_t *u8ptr = (uint8_t*)ptr;
    u8ptr += offset;

    *u8ptr = v;
}

void put_u16( void *ptr, size_t offset, uint16_t v )
{
    uint8_t *u8ptr = (uint8_t*)ptr;
    u8ptr += offset;

    std::uintptr_t p = reinterpret_cast< std::uintptr_t >( u8ptr );
    if ( ( p & 0x1 ) == 0 ) {
        // 16 bit aligned
        uint16_t *u16p = reinterpret_cast< uint16_t * >( p );
        *u16p = v;
        return;
    }

    memcpy( u8ptr, &v, sizeof( v ) );
}

void put_u32( void *ptr, size_t offset, uint32_t v )
{
    uint8_t *u8ptr = (uint8_t*)ptr;
    u8ptr += offset;

    std::uintptr_t p = reinterpret_cast< std::uintptr_t >( u8ptr );
    if ( ( p & 0x3 ) == 0 ) {
        // 32 bit aligned
        uint32_t *u32p = reinterpret_cast< uint32_t * >( p );
        *u32p = v;
        return;
    }

    memcpy( u8ptr, &v, sizeof( v ) );
}

void put_u64( void *ptr, size_t offset, uint64_t v )
{
    uint8_t *u8ptr = (uint8_t*)ptr;
    u8ptr += offset;

    std::uintptr_t p = reinterpret_cast< std::uintptr_t >( u8ptr );
    if ( ( p & 0x7 ) == 0 ) {
        // 64 bit aligned
        uint64_t *u64p = reinterpret_cast< uint64_t * >( p );
        *u64p = v;
        return;
    }

    memcpy( u8ptr, &v, sizeof( v ) );
}
