/* main.cc
 * Copyright (C) 2024 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cstdint>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include "lowlevel_types.hh"
#include "image.hh"
#include "palette.hh"
#include <algorithm>
#include "ab_logging.hh"
#include <getopt.h>
#include "options.hh"
#include <cassert>
#include "sdl_wrap.hh"
#include <future>
#include <list>
#include <chrono>
using namespace std::chrono_literals;
#include <unistd.h>
#include <climits>

#define DEFAULT_COLOR_TOLERANCE 0.15

static size_t progress_total_steps = 0;
static size_t progress_steps_done = 0;
static bool quiet = false;

class BinaryData {
public:
    void append_string( const std::string &str )
    {
        append_data( str.data(), str.size() );
    }

    void append_data( const std::vector< uint8_t > &vec )
    {
        m_data.insert( m_data.begin(),
                       vec.begin(),
                       vec.end() );
    }

    void append_data( const char *data, size_t data_len )
    {
        m_data.insert( m_data.end(),
                       data, data + data_len );
    }

    void append_data( const BinaryData &data )
    {
        m_data.insert( m_data.end(),
                       data.get_data().begin(),
                       data.get_data().end() );
    }

    const std::vector< uint8_t > &get_data() const
    {
        return m_data;
    }

    size_t size() const
    {
        return m_data.size();
    }
private:
    std::vector< uint8_t > m_data;
};

static std::string get_u32_as_buf( uint32_t v )
{
    char buf[128];

    put_u32( buf, 0, htobe_u32( v ) );

    return std::string( buf, 4 );
}

static BinaryData get_bmhd( const Image444 &img )
{
    BinaryData res;

    char buf[20] = {0};

    //m_width = betoh_u16( get_u16( buf, 0 ) );
    //m_height = betoh_u16( get_u16( buf, 2 ) );
    put_u16( buf, 0, htobe_u16( img.width() ) );
    put_u16( buf, 2, htobe_u16( img.height() ) );

    //m_pos_x = (int16_t)betoh_u16( get_u16( buf, 4 ) );
    //m_pos_y = (int16_t)betoh_u16( get_u16( buf, 6 ) );
    put_u16( buf, 4, htobe_u16( 0 ) );
    put_u16( buf, 6, htobe_u16( 0 ) );

    //m_no_of_planes = get_u8( buf, 8 );
    put_u8( buf, 8, 6 );

    //m_masking = get_u8( buf, 9 );
    put_u8( buf, 9, 0 );

    //m_compression = get_u8( buf, 10 );
    put_u8( buf, 10, global_options.compress_iff ? 1 : 0 );

    put_u8( buf, 11, 0 );
    //m_cmap_flags = get_u8( buf, 11 );

    //m_transparency_color = betoh_u16( get_u16( buf, 12 ) );
    put_u16( buf, 12, htobe_u16( 0 ) );

    //m_xaspect = get_u8( buf, 14 );
    //m_yaspect = get_u8( buf, 15 );
    put_u8( buf, 14, 1 );
    put_u8( buf, 15, 1 );

    //m_page_width = betoh_u16( get_u16( buf, 16 ) );
    //m_page_height = betoh_u16( get_u16( buf, 18 ) );
    put_u16( buf, 16, htobe_u16( img.width() ) );
    put_u16( buf, 18, htobe_u16( img.height() ) );

    res.append_data( buf, 20 );
    return res;
}

static BinaryData get_cmap( const Image444 &img )
{
    // colors * 3 byte (component << 4 | component )
    BinaryData res;

    const Palette &pal = img.get_palette();

    size_t colors = pal.size();

    if ( colors < 16 ) colors = 16;

    for ( size_t c = 0; c < colors; c++ ) {
        char buf[3];
        auto p = pal.get_color( c ).value_or( Pixel444() );

        put_u8( buf, 0, ( p.r << 4 ) | p.r );
        put_u8( buf, 1, ( p.g << 4 ) | p.g );
        put_u8( buf, 2, ( p.b << 4 ) | p.b );

        res.append_data( buf, 3 );
    }

    return res;
}

static BinaryData get_camg( const Image444 &img )
{
    BinaryData res;

    char buf[4] = {0};

    uint32_t mode;

    mode = 0x0800;  // HAM

    if ( global_options.interlace ) {
        mode |= 0x0004; // INTERLACE
    }

    put_u32( buf, 0, htobe_u32( mode ) );

    res.append_data( buf, 4 );

    return res;
}

static std::vector< uint8_t > convert_pixel_to_color_value( const Image444 &img )
{
    std::vector< uint8_t > res;

    for ( uint16_t y = 0; y < img.height(); y++ ) {
        for ( uint16_t x = 0; x < img.width(); x++ ) {
            auto pixel_res = img.get_ham_pixel( x, y );

            if ( ! pixel_res ) {
                return {};
            }

            auto & [ pixel, ham_value ] = *pixel_res;

            res.push_back( ham_value );
        }
    }

    return res;
}

static size_t same_block_length( const std::vector< uint8_t > &row,
                                 size_t start_offset )
{
    if ( start_offset >= row.size() ) return 0;

    uint8_t initial_value = row[ start_offset ];
    size_t count = 1;

    for ( size_t p = start_offset + 1; p < row.size(); p++, count++ ) {
        if ( row[ p ] != initial_value ) {
            break;
        }

        if ( count >= 128 ) break;
    }

    return count;
}

static size_t different_block_length( const std::vector< uint8_t > &row,
                                      size_t start_offset )
{
    if ( start_offset >= row.size() ) return 0;

    uint8_t prev_value = row[ start_offset ];
    uint8_t prev_prev_value = 0;
    size_t count = 1;

    for ( size_t p = start_offset + 1; p < row.size(); p++, count++ ) {
        if ( count >= 2 ) {
            if ( row[ p ] == prev_value &&
                 prev_prev_value == prev_value ) {
                count -= 2;
                break;
            }
        }

        prev_prev_value = prev_value;
        prev_value = row[ p ];

        if ( count >= 128 ) break;
    }

    return count;
}

static std::vector< uint8_t > apply_rle_compression( const std::vector< uint8_t > &row )
{
    std::vector< uint8_t > res;

    if ( ! global_options.compress_iff ) return row;

    for ( size_t o = 0; o < row.size(); ) {
        size_t repeat_count = same_block_length( row, o );

        if ( repeat_count > 1 ) {
            LOG_DEBUGXX( "same block " << repeat_count << " (from " << o << " to " << o + repeat_count << ")" );
            if ( repeat_count > 128 ) repeat_count = 128;

            uint8_t op = (uint8_t)( -( (int8_t)( repeat_count - 1 ) ) );
            res.push_back( op );
            res.push_back( row[o] );

            o += repeat_count;
            continue;
        }

        repeat_count = different_block_length( row, o );
        if ( repeat_count > 0 ) {
            LOG_DEBUGXX( "different block " << repeat_count << " (from " << o << " to " << o + repeat_count << ")" );
            if ( repeat_count > 128 ) repeat_count = 128;

            res.push_back( repeat_count - 1 );

            for ( size_t to = o; to < o + repeat_count; to++ ) {
                res.push_back( row[to] );
            }

            o += repeat_count;
            continue;
        }

        break;
    }

    // if a row must be 2-byte aligned, add a 0x80 nop here

    if ( res.size() > row.size() ) {
        LOG_DEBUGXX( "compressed is larger:" << res.size() << " vs " << row.size() );
    } else {
        LOG_DEBUGXX( "compressed to " << res.size() << " vs " << row.size() );
    }

    return res;
}

static std::vector< uint8_t > get_plane( size_t width,
                                         size_t y,
                                         const std::vector< uint8_t > &color_values,
                                         uint8_t plane )
{
    size_t row_bytes = ( ( width + 15 ) / 16 ) * 2;

    std::vector< uint8_t > res;

    uint8_t plane_value = 0;
    for ( size_t x = 0; x < width + 15 && res.size() < row_bytes; x++ ) {
        uint8_t bit = x & 0x7;
        uint8_t pixel = x < width ? color_values.at( y * width + x ) : 0;

        plane_value <<= 1;
        plane_value |= ( pixel & ( 1 << plane ) ) ? 1 : 0;

        if ( bit == 0x7 ) {
            res.push_back( plane_value );

            plane_value = 0;
        }
    }

    assert( res.size() == row_bytes );

    return apply_rle_compression( res );
}

static BinaryData get_body( const Image444 &img )
{
    BinaryData res;

    auto color_values = convert_pixel_to_color_value( img );

    for ( uint16_t y = 0; y < img.height(); y++ ) {
        for ( uint8_t p = 0; p < 6; p++ ) {
            auto plane = get_plane( img.width(), y, color_values, p );

            for ( size_t b = 0; b < plane.size(); b++ ) {
                char buf[1] = {(char)plane[b]};

                res.append_data( buf, 1 );
            }
        }
    }

    // make the result 2-byte aligned
    if ( ( res.size() & 1 ) == 1 ) {
        char buf[1] = {0};
        res.append_data( buf, 1 );
    }

    return res;
}

static BinaryData get_ilbm( const Image444 &img )
{
    auto bmhd = get_bmhd( img );
    auto cmap = get_cmap( img );
    auto camg = get_camg( img );
    auto body = get_body( img );

    BinaryData res;

    res.append_string( "BMHD" );
    size_t l = bmhd.size();
    res.append_string( get_u32_as_buf( l ) );
    res.append_data( bmhd );

    res.append_string( "CMAP" );
    l = cmap.size();
    res.append_string( get_u32_as_buf( l ) );
    res.append_data( cmap );

    res.append_string( "CAMG" );
    l = camg.size();
    res.append_string( get_u32_as_buf( l ) );
    res.append_data( camg );

    res.append_string( "BODY" );
    l = body.size();
    res.append_string( get_u32_as_buf( l ) );
    res.append_data( body );

    return res;
}

static BinaryData get_iff( const Image444 &img )
{
    BinaryData res;

    res.append_string( "FORM" );
    
    auto i = get_ilbm( img );
    size_t l = i.size() + 4;
    res.append_string( get_u32_as_buf( l ) );
    res.append_string( "ILBM" );
    res.append_data( i );

    return res;
}

class color_iterator {
public:
    color_iterator() = delete;
    color_iterator( const std::vector< Pixel444 > &color_pool ) : m_color_pool( color_pool )
    {}

    std::optional< Pixel444 > get_next()
    {
        const std::lock_guard< std::mutex > lock( cols_mutex );

        if ( ! m_color_pool.empty() ) {
            if ( m_color_pool_pos >= m_color_pool.size() ) return {};
            return m_color_pool[ m_color_pool_pos++ ];
        }

        if ( m_done ) return {};

        Pixel444 res( m_r, m_g, m_b );

        if ( ++m_b >= 16 ) {
            m_b = 0;
            if ( ++m_g >= 16 ) {
                m_g = 0;
                if ( ++m_r >= 16 ) {
                    m_r = 0;
                }
            }
        }

        if ( m_r == 0 && m_g == 0 && m_b == 0 ) m_done = true;
        
        return res;
    }

    double progress() const
    {
        // dont care about race conditions here

        if ( ! m_color_pool.empty() ) {
            return (double)m_color_pool_pos / m_color_pool.size();
        }

        size_t t = m_r;
        t *= 16;
        t += m_g;
        t *= 16;
        t += m_b;
        double r = (double)t / ( 16 * 16 * 16 );

        return r;
    }
private:
    const std::vector< Pixel444 > &m_color_pool;
    size_t m_color_pool_pos = 0;
    uint8_t m_r = 0;
    uint8_t m_g = 0;
    uint8_t m_b = 0;
    bool m_done = false;
    std::mutex cols_mutex;
};

static std::vector< Pixel444 > find_best_palette( const Image444 &img,
                                                  const std::vector< Pixel444 > &initial_colors,
                                                  const std::vector< Pixel444 > &color_pool,
                                                  size_t color_pos,
                                                  size_t color_limit )
{
    if ( color_pos >= color_limit ) {
        LOG_DEBUGXX( "  recursion end" );
        return initial_colors;
    }

    color_iterator cols( color_pool );

    std::vector< Pixel444 > best_pal;
    double best_error = 0;
    bool best_set = false;
    Pixel444 best_col;

    LOG_DEBUGXX( "called for pos " << color_pos );

    time_t t1 = time( NULL );

    std::list< std::future< std::tuple< bool, double, Pixel444 > > > results;
    for ( size_t t = 0; t < global_options.threads; t++ ) {
        auto f = std::async( std::launch::async,
                             [ &img,
                               &initial_colors,
                               &cols,
                               color_pos ] () -> std::tuple< bool, double, Pixel444 > {
                                 double best_error = 0;
                                 bool best_set = false;
                                 Pixel444 best_col;

                                 while ( auto p = cols.get_next() ) {
                                     std::vector< Pixel444 > new_pal = initial_colors;

                                     if ( new_pal.size() <= color_pos ) {
                                         new_pal.resize( color_pos + 1 );
                                     }
                                     new_pal[color_pos] = *p;
        
                                     Image444 new_img( img );
                                     Palette pal;
                                     for ( const auto &e : new_pal ) {
                                         pal.add_color( e );
                                     }
                                     new_img.reset_marks();
                                     new_img.set_palette( pal );
                                     new_img.simple_hamify();

                                     double e = new_img.calc_error( img );

                                     if ( ! best_set || e < best_error ) {
                                         best_error = e;
                                         best_set = true;
                                         best_col = *p;
                                     }
                                 }

                                 return { best_set, best_error, best_col };
                             });
        results.push_back( std::move( f ) );
    }

    for ( auto &e : results ) {
        for (;;) {
            std::future_status status = e.wait_for(100ms);
            if ( status == std::future_status::ready ) break;

            if ( ! quiet ) {
                fprintf( stdout, "Progress: %6.2f%%\r", ( (double)( progress_steps_done + color_pos ) * 100 / progress_total_steps ) + ( cols.progress() * 100 / progress_total_steps ) );
                fflush( stdout );
            }
        }
        
        auto [ set, err, col ] = e.get();
        if ( ! set ) continue;
        
        if ( ! best_set || err < best_error ) {
            best_error = err;
            best_set = true;
            best_col = col;
        }
    }

    time_t t2 = time( NULL );
    LOG_DEBUGXX( "  step time: " << t2 - t1 );
    LOG_DEBUGXX( "  best color found: " << best_col );
    LOG_DEBUGXX( "  with error " << best_error );

    best_pal = initial_colors;

    if ( best_pal.size() <= color_pos ) {
        best_pal.resize( color_pos + 1 );
    }
    best_pal[color_pos] = best_col;

    return find_best_palette( img, best_pal, color_pool, color_pos + 1, color_limit );
}

enum {
    OPT_DEBUG_FILE = 257,
    OPT_DEBUG_FUNCTION,
    OPT_SCALE,
    OPT_INTERLACE,
    OPT_DITHER,
    OPT_NO_DITHER,
    OPT_COMPRESS,
    OPT_NTSC,
    OPT_QUIET,
    OPT_THREADS,
    OPT_ADJUST_INTERLACE,
};

static void show_help()
{
    printf( "hamify: ILBM HAM6 converter\n");
    printf( "  Version: %s\n",PACKAGE_VERSION);
    printf( "  Author: Ralf Hoffmann <ralf@boomerangsworld.de>\n" );
    printf( "  Homepage: http://www.boomerangsworld.de/cms/tools/hamify\n\n" );
    printf( "  hamify [options] <file name>\n" );
    printf( "  Options: -V, --version\tPrint the program version and exit\n" );
    printf( "           -h, --help\t\tPrint this help and exit\n" );
    printf( "           -o, --output=<path>\tOutput file\n" );
    printf( "           -t, --type=<type>\tOutput type\n" );
    printf( "\t\t\t\tPossible values: iff, ppm\n" );
    printf( "           --scale\t\tScale image to OCS LoRes\n" );
    printf( "           --interlace\t\tUse interlace screen size\n" );
    printf( "           --adjust-interlace\tWhen used without scaling, still adjust the\n" );
    printf( "\t\t\t\twidth so the aspect ratio is correct on\n" );
    printf( "\t\t\t\tinterlace screen.\n" );
    printf( "           --ntsc\t\tSelect NTSC screen size\n" );
    printf( "           --no-dither\t\tDisable dithering when converting to OCS color\n" );
    printf( "\t\t\t\tspace.\n" );
    printf( "           -q, --quality\tSelect quality (speed vs accuracy)\n" );
    printf( "\t\t\t\tPossible values: 0 or fast\n" );
    printf( "\t\t\t\t                 1 or normal (default)\n" );
    printf( "\t\t\t\t                 2 or good, 3 or best\n" );
    printf( "\t\t\t\tfast and normal are reasonable fast, while\n" );
    printf( "\t\t\t\tgood or best take signifcant amount of time\n" );
    printf( "\t\t\t\tto find most accurat representation.\n" );
    printf( "           -l, --ham-lookahead\tSelect type of lookahead mode for HAM\n" );
    printf( "\t\t\t\tcalculation.\n" );
    printf( "\t\t\t\tPossible values: 0 for fast calculation\n" );
    printf( "\t\t\t\t                 1 use HAM step with lowest\n" );
    printf( "\t\t\t\t                   error\n" );
    printf( "\t\t\t\t                 2 use one lookahead pixel to\n" );
    printf( "\t\t\t\t                   choose best HAM step\n" );
    printf( "\t\t\t\tDefault value: 2\n" );
    printf( "           --compress\t\tEnable ILBM compression\n" );
    printf( "           --quiet\t\tSuppress regular output\n" );
    printf( "           --threads=<threads>\tUse given number of threads for faster\n" );
    printf( "\t\t\t\tprocessing in good and best quality mode.\n" );
    printf( "\t\t\t\tDefaults to number of threads available on\n" );
    printf( "\t\t\t\tsystem.\n" );
    printf( "           -d, --debug-level=<level>\n" );
    printf( "\t\t\t\tSet debug level (0=none, 1=critical, 2=verbose,\n" );
    printf( "\t\t\t\t3=debug)\n" );
    printf( "           --debug-file=<file>\tShow debug message only from file\n" );
    printf( "           --debug-function=<function>\n" );
    printf( "\t\t\t\tShow debug message only from function\n" );
    printf( "           <input file>\tProcess given input file\n\n" );
}

static void show_version()
{
    printf( "%s\n", PACKAGE_STRING );
    printf( "  Author: Ralf Hoffmann <ralf@boomerangsworld.de>\n" );
    printf( "  Homepage: http://www.boomerangsworld.de/cms/tools/hamify\n\n" );
}

static void check_options( int argc, char **argv )
{
    while ( true ) {
        int option_index;
        static struct option long_options[] = {
                                               { "help", no_argument, NULL, 'h' },
                                               { "version", no_argument, NULL, 'V' },
                                               { "scale", no_argument, NULL, OPT_SCALE },
                                               { "interlace", no_argument, NULL, OPT_INTERLACE },
                                               { "adjust-interlace", no_argument, NULL, OPT_ADJUST_INTERLACE },
                                               { "dither", no_argument, NULL, OPT_DITHER },
                                               { "no-dither", no_argument, NULL, OPT_NO_DITHER },
                                               { "ham-lookahead", required_argument, NULL, 'l' },
                                               { "quality", required_argument, NULL, 'q' },
                                               { "type", required_argument, NULL, 't' },
                                               { "output-file", required_argument, NULL, 'o' },
                                               { "compress", no_argument, NULL, OPT_COMPRESS },
                                               { "ntsc", no_argument, NULL, OPT_NTSC },
                                               { "quiet", no_argument, NULL, OPT_QUIET },
                                               { "threads", required_argument, NULL, OPT_THREADS },
                                               { "debug-level", required_argument, NULL, 'd' },
                                               { "debug-file", required_argument, NULL, OPT_DEBUG_FILE },
                                               { "debug-function", required_argument, NULL, OPT_DEBUG_FUNCTION },
                                               { NULL, 0, NULL, 0 }
        };

        int c = getopt_long( argc, argv, "hVl:d:q:o:t:", long_options, &option_index );

        if ( c == -1 ) break;
      
        switch( c ) {
             case 'h':
                show_help();
                exit(0);
                break;
            case 'V':
                show_version();
                exit(0);
                break;
            case 'o':
                if ( optarg != NULL ) {
                    global_options.output_file = optarg;
                } else {
                    LOG_CRITICALXX( "Output file missing" );
                    exit( 1 );
                }
                break;
            case 'l':
                global_options.lookahead_mode = ham_options::DIRECT_LOOKAHEAD;
                if ( optarg != NULL ) {
                    std::string value = optarg;

                    if ( value == "0" ) {
                        global_options.lookahead_mode = ham_options::NO_LOOKAHEAD_FAST;
                    } else if ( value == "1" ) {
                        global_options.lookahead_mode = ham_options::NO_LOOKAHEAD_NORMAL;
                    } else if ( value == "2" ) {
                        global_options.lookahead_mode = ham_options::DIRECT_LOOKAHEAD;
                    } else {
                        LOG_CRITICALXX( "Lookahead mode '" << value << " not supported (only {0,1,2})" );
                        exit( 1 );
                    }
                }
                break;
            case 'q':
                global_options.quality_mode = ham_options::QUALITY_NORMAL;
                if ( optarg != NULL ) {
                    std::string value = optarg;

                    if ( value == "0" || value == "fast" ) {
                        global_options.quality_mode = ham_options::QUALITY_FAST;
                    } else if ( value == "1" || value == "normal" ) {
                        global_options.quality_mode = ham_options::QUALITY_NORMAL;
                    } else if ( value == "2" || value == "good" ) {
                        global_options.quality_mode = ham_options::QUALITY_GOOD;
                    } else if ( value == "3" || value == "best" ) {
                        global_options.quality_mode = ham_options::QUALITY_BEST;
                    } else {
                        LOG_CRITICALXX( "quality mode '" << value << " not supported (only {0,1,2,3})" );
                        exit( 1 );
                    }
                }
                break;
            case 't':
                global_options.output_type = ham_options::OUTPUT_IFF;
                if ( optarg != NULL ) {
                    std::string value = optarg;

                    if ( value == "ppm" ) {
                        global_options.output_type = ham_options::OUTPUT_PPM;
                    } else if ( value == "iff" ) {
                        global_options.output_type = ham_options::OUTPUT_IFF;
                    } else {
                        LOG_CRITICALXX( "type '" << value << " not supported (only {iff,ppm})" );
                        exit( 1 );
                    }
                }
                break;
            case OPT_THREADS:
                if ( optarg == NULL ) {
                    show_help();
                    exit(1);
                } else {
                    int threads = atoi( optarg );
                    if ( threads < 1 ) {
                        global_options.threads = 1;
                    } else {
                        global_options.threads = threads;
                    }
                }
                break;
            case 'd':
                if ( optarg == NULL ) {
                    show_help();
                    exit(1);
                } else {
                    int level = atoi( optarg );
                    switch ( level ) {
                        case 0:
                            ab_set_log_level( LOG_NOTHING );
                            break;
                        case 1:
                            ab_set_log_level( LOG_CRITICAL );
                            break;
                        case 2:
                            ab_set_log_level( LOG_VERBOSE );
                            break;
                        case 3:
                            ab_set_log_level( LOG_DEBUG );
                            break;
                        default:
                            show_help();
                            exit(1);
                    }
                }
                break;
            case OPT_DEBUG_FILE:
                if ( ! optarg ) {
                    show_help();
                    exit(1);
                } else {
                    ab_add_file_filter( optarg );
                }
                break;
            case OPT_DEBUG_FUNCTION:
                if ( ! optarg ) {
                    show_help();
                    exit(1);
                } else {
                    ab_add_function_filter( optarg );
                }
                break;
            case OPT_SCALE:
                global_options.scale = true;
                break;
            case OPT_INTERLACE:
                global_options.interlace = true;
                break;
            case OPT_ADJUST_INTERLACE:
                global_options.adjust_interlace = true;
                break;
            case OPT_DITHER:
                global_options.dither = true;
                break;
            case OPT_NO_DITHER:
                global_options.dither = false;
                break;
            case OPT_COMPRESS:
                global_options.compress_iff = true;
                break;
            case OPT_NTSC:
                global_options.ntsc = true;
                break;
            case OPT_QUIET:
                quiet = true;
                break;
            default:
                break;
        }
    }

    if ( optind < argc ) {
        global_options.input_file = argv[ optind ];
    } else {
        LOG_CRITICALXX( "Input file missing" );
        exit( 1 );
    }
}

int main( int argc, char **argv )
{
    ab_set_log_level( LOG_CRITICAL );

    global_options.lookahead_mode = ham_options::DIRECT_LOOKAHEAD;
    global_options.quality_mode = ham_options::QUALITY_NORMAL;

#ifdef _SC_NPROCESSORS_ONLN
    long r = sysconf( _SC_NPROCESSORS_ONLN );
    if ( r > 0 ) {
        global_options.threads = r;
    }
#endif

    check_options( argc, argv );

    LOG_VERBOSEXX( "Options:" );
    LOG_VERBOSEXX( "  Lookahead mode: " << global_options.lookahead_mode );
    LOG_VERBOSEXX( "  Quality mode: " << global_options.quality_mode );
    LOG_VERBOSEXX( "  Scale to screen size: " << global_options.scale );
    LOG_VERBOSEXX( "  Interlace: " << global_options.interlace );
    LOG_VERBOSEXX( "  Adjust for interlace: " << global_options.adjust_interlace );
    LOG_VERBOSEXX( "  Dithering: " << global_options.dither );
    LOG_VERBOSEXX( "  Threads: " << global_options.threads );

    // fast: unique -> best_pal( default tolerance)
    // normal: unique -> find best tolerance
    // good: unique -> find best palette from unique colors
    // best: unique -> find best palette from complete color space

    SDL_Wrap sdl;
    
    ImageFP original_img;
    int res = original_img.read( global_options.input_file );

    if ( res != 0 ) {
        std::cout << "Failed to read image '" << global_options.input_file << "'" << std::endl;
        return 1;
    }

    if ( global_options.scale ) {
        size_t screen_w = 320;
        size_t screen_h = global_options.ntsc ? 240 : 256;

        double sx = (double)original_img.width() / screen_w;
        double sy = (double)original_img.height() / screen_h;
        double scale;
        if ( sx > sy ) {
            scale = sx;
        } else {
            scale = sy;
        }

        size_t scaled_w = (double)original_img.width() / scale + 0.5;
        size_t scaled_h = (double)original_img.height() * ( global_options.interlace ? 2 : 1 ) / scale + 0.5;

        LOG_VERBOSEXX( "scale from " << original_img.width() << "x" << original_img.height() << " to " << scaled_w << "x" << scaled_h );

        original_img = original_img.scale( scaled_w, scaled_h );
    } else  if ( global_options.interlace && global_options.adjust_interlace ) {
        original_img = original_img.scale( original_img.width() / 2,
                                           original_img.height() );
    }

    if ( global_options.dither ) {
        original_img = original_img.dither_444();
    }

    auto img = original_img.convert_to_444();

    LOG_VERBOSEXX( "Got image dimensions " << img.width() << " x " << img.height() );
    LOG_VERBOSEXX( "Unique colors OCS: " << img.calc_unique_colors() );

    auto unique_colors = img.get_unique_colors( false );

    progress_total_steps = 1;
    if ( global_options.quality_mode >= ham_options::QUALITY_GOOD ) {
        progress_total_steps += 16;
    }

    std::vector< Pixel444 > selected_colors;
    if ( global_options.quality_mode == ham_options::QUALITY_FAST ) {
        selected_colors = img.calc_palette( 16, DEFAULT_COLOR_TOLERANCE );

        if ( selected_colors.size() < 16 && unique_colors.size() > selected_colors.size() ) {
            LOG_DEBUGXX( "fill up remaining color slots" );

            std::vector< Pixel444 > new_colors = selected_colors;

            for ( auto &e: unique_colors ) {
                if ( new_colors.size() >= 16 ) break;
                bool existing = false;
                for ( auto &c: new_colors ) {
                    if ( e.first == c ) {
                        existing = true;
                        break;
                    }
                }
                if ( ! existing ) {
                    new_colors.push_back( e.first );
                }
            }

            selected_colors = new_colors;
        }

        LOG_DEBUGXX( "use palette for default tolerance" );
        for ( auto &c : selected_colors ) {
            LOG_DEBUGXX( "  " << c );
        }
    } else {
        Palette best_pal;
        double best_err = std::numeric_limits<double>::infinity();
        double best_tol = 0;
        auto unique_colors = img.get_unique_colors( true );

        if ( ! unique_colors.empty() ) {
            for ( double tol = 0.001; tol <= 1.0; tol += 0.005 ) {
                LOG_DEBUGXX( "check tolerance " << tol );

                if ( ! quiet ) {
                    fprintf( stdout, "Progress: %6.2f%%\r", ( tol * 100 / progress_total_steps ) );
                    fflush( stdout );
                }

                img.reset_marks();
                auto alt_pal = img.calc_palette( unique_colors, 16, tol );

                Palette pal_alt;
                for ( const auto &e : alt_pal ) {
                    pal_alt.add_color( e );
                }

                if ( alt_pal.size() < 16 && best_pal.size() > 0 ) break;
                if ( pal_alt == best_pal ) continue;

                LOG_DEBUGXX( "palette changed, check error" );

                Image444 ham_img( img );
                ham_img.reset_marks();
                ham_img.set_palette( pal_alt );
                ham_img.simple_hamify();

                double new_err = ham_img.calc_error( original_img );

                if ( new_err < best_err ) {
                    best_err = new_err;
                    best_pal = pal_alt;
                    best_tol = tol;

                    LOG_DEBUGXX( "found new best palette for tolerance " << tol );
                }
            }
        }

        LOG_DEBUGXX( "found best palette for tolerance " << best_tol );
        for ( size_t c = 0; c < best_pal.size(); c++ ) {
            auto e = best_pal.get_color( c );
            LOG_DEBUGXX( "  " << *e );
            selected_colors.push_back( *e );
        }
    }

    progress_steps_done++;

    if ( global_options.quality_mode >= ham_options::QUALITY_GOOD ) {
        std::vector< Pixel444 > all_colors;
        std::vector< Pixel444 > start_colors = selected_colors;

        for ( auto &e: unique_colors ) {
            all_colors.push_back( e.first );
        }

        if ( global_options.quality_mode == ham_options::QUALITY_GOOD ) {
            auto colors = find_best_palette( img, start_colors, all_colors, 0, 16 );
            selected_colors = colors;
        } else {
            auto colors = find_best_palette( img, start_colors, {}, 0, 16 );
            selected_colors = colors;
        }
    }

    if ( ! quiet ) {
        fprintf( stdout, "Progress: 100.00%%\n" );
    }

    Palette best_pal;
    for ( const auto &e : selected_colors ) {
        best_pal.add_color( e );
    }
    Image444 ham_img( img );
    ham_img.set_palette( best_pal );
    ham_img.simple_hamify();

    LOG_VERBOSEXX( "HAM image error: " << ham_img.calc_error( original_img ) );

    if ( global_options.output_file.empty() ) {
        std::cout << "No output file name given" << std::endl;
        return 1;
    } else {
        if ( global_options.output_type == ham_options::OUTPUT_PPM ) {
            bool res = ham_img.write_ppm( global_options.output_file );

            if ( ! res ) {
                std::cout << "Failed to write ppm file '" << global_options.output_file << "'" << std::endl;
                return 1;
            }
        } else {
            std::ofstream ofile( global_options.output_file, std::ios::binary );
            auto iff_data = get_iff( ham_img );

            ofile.write( reinterpret_cast<const char*>( &iff_data.get_data()[0] ), iff_data.size() );

            ofile.close();
            if ( ! ofile.good() ) {
                std::cout << "failed to write IFF file '" << global_options.output_file << "'" << std::endl;
                return 1;
            }
        }
    }

    return 0;
}
