/* palette.cc
 * Copyright (C) 2024 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "palette.hh"
#include <cerrno>

Palette::Palette( size_t colors )
{
    m_colors.resize( colors );
}

std::optional< Pixel444 > Palette::get_color( size_t color ) const
{
    if ( color >= m_colors.size() ) return {};

    return m_colors[ color ];
}

int Palette::set_color( size_t color, Pixel444 v )
{
    if ( color >= m_colors.size() ) return -ERANGE;

    m_colors[ color ] = v;

    update_nearest_lookup();

    return 0;
}

int Palette::add_color( Pixel444 v )
{
    m_colors.push_back( v );

    update_nearest_lookup();

    return 0;
}

std::optional< size_t > Palette::find_color( Pixel444 v ) const
{
    size_t pos = ( v.r << 8 ) | ( v.g << 4 ) | v.b;

    if ( pos < m_nearest_lookup.size() ) {
        if ( m_nearest_lookup[ pos ].second == 2 ) {
            return m_nearest_lookup[pos].first;
        } else {
            return {};
        }
    }

    for ( size_t p = 0; p < m_colors.size(); p++ ) {
        if ( m_colors.at( p ) == v ) return p;
    }

    return {};
}

size_t Palette::size() const
{
    return m_colors.size();
}

std::optional< size_t > Palette::find_nearest_color( Pixel444 v )
{
    if ( m_colors.empty() ) return {};

    size_t pos = ( v.r << 8 ) | ( v.g << 4 ) | v.b;

    if ( pos >= m_nearest_lookup.size() ) {
        m_nearest_lookup.resize( pos + 1 );
    } else if ( m_nearest_lookup[ pos ].second > 0 ) {
        return m_nearest_lookup[pos].first;
    }
    
    size_t nearest = m_colors.size();
    double nearest_distance = 0;
    for ( size_t p = 0; p < m_colors.size(); p++ ) {
        if ( p == 0 ) {
            nearest = 0;
            nearest_distance = color_distance( v, m_colors[p] );
            continue;
        }

        if ( color_distance( v, m_colors[p] ) < nearest_distance ) {
            nearest_distance = color_distance( v, m_colors[p] );
            nearest = p;
        }
    }

    m_nearest_lookup[pos] = { nearest, 1 };
    
    return nearest;
}

double Palette::color_distance( const Pixel444 &p1,
                                const Pixel444 &p2 )
{
    return sqrt( ( (double)p1.r - p2.r ) * ( (double)p1.r - p2.r ) +
                 ( (double)p1.g - p2.g ) * ( (double)p1.g - p2.g ) +
                 ( (double)p1.b - p2.b ) * ( (double)p1.b - p2.b ) );
}

bool Palette::operator==( const Palette &oth ) const
{
    return m_colors == oth.m_colors;
}

void Palette::update_nearest_lookup()
{
    m_nearest_lookup.clear();

    for ( size_t p = 0; p < m_colors.size(); p++ ) {
        auto &c = m_colors[p];
        size_t pos = ( c.r << 8 ) | ( c.g << 4 ) | c.b;

        if ( pos >= m_nearest_lookup.size() ) {
            m_nearest_lookup.resize( pos + 1 );
        }

        m_nearest_lookup[ pos ] = { p, 2 };
    }
}
