/* pixel.hh
 * Copyright (C) 2024 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PIXEL_HH
#define PIXEL_HH

#include <cstdint>
#include <cstring>
#include <iostream>
#include <cmath>
#include <optional>
#include "ab_logging.hh"

struct ocs_pixel {
    uint32_t r:4;
    uint32_t g:4;
    uint32_t b:4;
};

struct fp_pixel {
    float r;
    float g;
    float b;
};

struct ocs_pixel_palette {
    uint8_t pal_entry:6;
};

enum pixel_type {
    PIXEL_FLOAT,
    PIXEL_OCS,
    PIXEL_OCS_PALETTE
};

class Pixel444;

class PixelFP {
public:
    PixelFP() : r( 0.0 ),
                g( 0.0 ),
                b( 0.0 ),
                a( 1.0 )
    {}
    explicit PixelFP( double r, double g, double b, double a = 1.0 );
    explicit PixelFP( uint32_t r, uint32_t g, uint32_t b, uint32_t a, uint32_t max_value );
    
    double r;
    double g;
    double b;
    double a;

    Pixel444 convert() const;

    friend std::ostream& operator<<( std::ostream& stream, const PixelFP &p )
    {
        stream << "(" << p.r << "," << p.g << "," << p.b << "," << p.a << ")";
        return stream;
    }

    double distance( const PixelFP &oth ) const
    {
        return sqrt( ( r - oth.r ) * ( r - oth.r ) +
                     ( g - oth.g ) * ( g - oth.g ) +
                     ( b - oth.b ) * ( b - oth.b ) );
    }

};

class Pixel444 {
public:
    Pixel444() : r( 0 ), g( 0 ), b( 0 )
    {}
    Pixel444( const Pixel444 &oth ) = default;
    Pixel444( uint16_t r, uint16_t g, uint16_t b );
    
    PixelFP convert() const;

    bool operator==( const Pixel444 &oth ) const
    {
        if ( r != oth.r ) return false;
        if ( g != oth.g ) return false;
        if ( b != oth.b ) return false;
        return true;
    }

    bool operator<( const Pixel444 &oth ) const
    {
        if ( r < oth.r ) return true;
        if ( r > oth.r ) return false;

        if ( g < oth.g ) return true;
        if ( g > oth.g ) return false;

        if ( b < oth.b ) return true;
        return false;
    }

    double distance( const Pixel444 &oth ) const
    {
        return sqrt( ( (double)r - oth.r ) * ( (double)r - oth.r ) +
                     ( (double)g - oth.g ) * ( (double)g - oth.g ) +
                     ( (double)b - oth.b ) * ( (double)b - oth.b ) );
    }

    friend std::ostream& operator<<( std::ostream& stream, const Pixel444 &p )
    {
        stream << "(" << p.r << "," << p.g << "," << p.b << ")";
        return stream;
    }

    uint16_t r:4;
    uint16_t g:4;
    uint16_t b:4;
};

#endif /* PIXEL_HH */
